# -*- coding: utf-8 -*-
"""
author: steve simmert
bugs and hints to: steve.simmert@uni-tuebingen.de
"""

names = 'axis_names'

laterality = 'laterality'

height = 'height'  # sample height
height_unit = 'height_unit'

N_avg = 'number_of_averages'
f_sample = 'sampling_rate'

density = density_p = 'density_particle'
density_p_unit = 'density_particle_unit'

density_m = 'density_medium'
density_m_unit = 'density_medium_unit'

viscosity = 'viscosity'
viscosity_unit = 'viscosity_unit'

material = 'material'

medium = 'medium'

temp = 'temperature'
temp_err = 'temperature_err'
temp_unit = temp_unit = 'temperature_unit'

bead_dia = 'bead_diameter'
bead_dia_err = 'bead_diameter_err'
bead_dia_unit = 'bead_diameter_unit'

radius = 'radius'
radius_err = 'radius_err'
radius_unit = 'radius_unit'

freq = 'freq'
freq_unit = 'frequency_unit'

PSD_x = 'PSD_x'
PSD_y = 'PSD_y'
PSD_z = 'PSD_z'
psd_unit = 'psd_unit'

ex_axis = 'excitation_axis'
ex_freq = 'excitation_frequency'
ex_amp = 'excitation_amplitude'
ex_amp_err = 'excitation_amplitude_err'
ex_amp_unit = 'excitation_amplitude_unit'
ex_pow = 'reference power'
ex_pow_err = 'reference_power_err'
ex_pow_unit = 'reference_power_unit'

z_signal = 'z_signal'

wl = 'wavelength'  # optical wavelength (in air) of the trapping laser
wl_unit = 'wavelength_unit'

fs = focal_shift = 'focal_shift'
dfs = 'focal_shift_err'

ref_ind = 'refractive_index'  # refractive index of the medium

conf_level = 'confidence_level'

fit_method = 'fit_method'

corr = 'correction_factor'
corr_err = 'correction_factor_error'

d_app_surf = 'apparent_surface_distance'
d_app_surf_err = 'apparent_surface_distance_err'
h0 = 'h0'
h0_err = 'h0_err'
height_offset = 'applied_height_offset'

dissens = 'displacement_sensitivity'
dissens_err = 'displacement_sensitivity_error'
dissens_unit = beta_unit = 'displacement_sensitivity_unit'

trap_stiffness = 'trap_stiffness'
trap_stiffness_err = 'trap_stiffness_error'
trap_stiffness_unit = 'trap_stiffness_unit'
kappa_unit = 'trapstiffness_unit'

drag = 'drag_coefficient'
drag_err = 'drag_error'
drag_unit = 'drag_unit'

f_c = 'corner_frequency'
df_c = 'corner_frequency_err'
f_c_unit = 'corner_frequency_unit'

D = 'Diffusion_coefficient'
dD = 'Diffusion_coefficient_err'
D_unit = 'Diffusion_coefficient_unit'

model = 'model_name'
redchi2 = 'reduced_chi2'
nfree = 'number_degrees_freedom'
outlier = 'outlier'
conf_level = 'confidence_level'
bounds = 'freq_bounds'
